@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 
 # =============================================================================
-#  Version: 3.0 (July 22, 2020)
+#  Version: 3.0 (January 24, 2023)
 #  Author: Giuseppe Attardi (attardi@di.unipi.it), University of Pisa
 #
 #  Contributors:
@@ -17,7 +17,7 @@
 #   Nick Ulven (nulven@github)
 #
 # =============================================================================
-#  Copyright (c) 2009-2020. Giuseppe Attardi (attardi@di.unipi.it).
+#  Copyright (c) 2009-2023. Giuseppe Attardi (attardi@di.unipi.it).
 # =============================================================================
 #  This file is part of Tanl.
 #
@@ -43,7 +43,13 @@ Each file will contain several documents in the format:
         ...
         </doc>
 
-This version performs template expansion by preprocesssng the whole dump and
+If the program is invoked with the --json flag, then each file will                                            
+contain several documents formatted as json ojects, one per line, with                                         
+the following structure
+
+    {"id": "", "revid": "", "url": "", "title": "", "text": "..."}
+
+The program performs template expansion by preprocesssng the whole dump and
 collecting template definitions.
 """
 
@@ -54,7 +60,7 @@ import os.path
 import re  # TODO use regex when it will be standard
 import sys
 from io import StringIO
-from multiprocessing import Queue, Process, cpu_count
+from multiprocessing import Queue, get_context, cpu_count
 from timeit import default_timer
 
 from .extract import Extractor, ignoreTag, define_template, acceptedNamespaces
@@ -62,7 +68,7 @@ from .extract import Extractor, ignoreTag, define_template, acceptedNamespaces
 # ===========================================================================
 
 # Program version
-__version__ = '3.0.4'
+__version__ = '3.0.7'
 
 ##
 # Defined in <siteinfo>
@@ -73,17 +79,12 @@ knownNamespaces = set(['Template'])
 # The namespace used for template definitions
 # It is the name associated with namespace key=10 in the siteinfo header.
 templateNamespace = ''
-templatePrefix = ''
 
 ##
 # The namespace used for module definitions
 # It is the name associated with namespace key=828 in the siteinfo header.
 moduleNamespace = ''
 
-# This is obtained from <siteinfo>
-urlbase = ''
-
-
 # ----------------------------------------------------------------------
 # Modules
 
@@ -101,7 +102,7 @@ modules = {
 
 # def expandTemplates(text):
 #     """Expand templates invoking MediaWiki API"""
-#     text = urlib.urlencodew(text.encode('utf-8'))
+#     text = urlib.urlencodew(text)
 #     base = urlbase[:urlbase.rfind('/')]
 #     url = base + "/w/api.php?action=expandtemplates&format=json&text=" + text
 #     exp = json.loads(urllib.urlopen(url))
@@ -167,7 +168,10 @@ class OutputSplitter():
 
     def write(self, data):
         self.reserve(len(data))
-        self.file.write(data)
+        if self.compress:
+            self.file.write(data)
+        else:
+            self.file.write(data)
 
     def close(self):
         self.file.close()
@@ -190,9 +194,9 @@ def load_templates(file, output_file=None):
     """
     Load templates from :param file:.
     :param output_file: file where to save templates and modules.
+    :return: number of templates loaded.
     """
-    global templateNamespace, templatePrefix
-    templatePrefix = templateNamespace + ':'
+    global templateNamespace
     global moduleNamespace, modulePrefix
     modulePrefix = moduleNamespace + ':'
     articles = 0
@@ -215,6 +219,13 @@ def load_templates(file, output_file=None):
             page = []
         elif tag == 'title':
             title = m.group(3)
+            if not output_file and not templateNamespace:  # do not know it yet
+                # we reconstruct it from the first title
+                colon = title.find(':')
+                if colon > 1:
+                    templateNamespace = title[:colon]
+                    Extractor.templatePrefix = title[:colon + 1]
+            # FIXME: should reconstruct also moduleNamespace
         elif tag == 'text':
             inText = True
             line = line[m.start(3):m.end(3)]
@@ -228,25 +239,18 @@ def load_templates(file, output_file=None):
         elif inText:
             page.append(line)
         elif tag == '/page':
-            if not output_file and not templateNamespace:  # do not know it yet
-                # we reconstruct it from the first title
-                colon = title.find(':')
-                if colon > 1:
-                    templateNamespace = title[:colon]
-                    templatePrefix = title[:colon + 1]
-            # FIXME: should reconstruct also moduleNamespace
-            if title.startswith(templatePrefix):
+            if title.startswith(Extractor.templatePrefix):
                 define_template(title, page)
                 templates += 1
             # save templates and modules to file
-            if output_file and (title.startswith(templatePrefix) or
+            if output_file and (title.startswith(Extractor.templatePrefix) or
                                 title.startswith(modulePrefix)):
                 output.write('<page>\n')
-                output.write('   <title>%s</title>\n' % title.encode('utf-8'))
+                output.write('   <title>%s</title>\n' % title)
                 output.write('   <ns>10</ns>\n')
                 output.write('   <text>')
                 for line in page:
-                    output.write(line.encode('utf-8'))
+                    output.write(line)
                 output.write('   </text>\n')
                 output.write('</page>\n')
             page = []
@@ -262,6 +266,7 @@ def load_templates(file, output_file=None):
 def decode_open(filename, mode='rt', encoding='utf-8'):
     """
     Open a file, decode and decompress, depending on extension `gz`, or 'bz2`.
+    :param filename: the file to open.
     """
     ext = os.path.splitext(filename)[1]
     if ext == '.gz':
@@ -273,8 +278,65 @@ def decode_open(filename, mode='rt', encoding='utf-8'):
         return open(filename, mode, encoding=encoding)
 
 
+def collect_pages(text):
+    """
+    :param text: the text of a wikipedia file dump.
+    """
+    # we collect individual lines, since str.join() is significantly faster
+    # than concatenation
+    page = []
+    id = ''
+    revid = ''
+    last_id = ''
+    inText = False
+    redirect = False
+    for line in text:
+        if '<' not in line:     # faster than doing re.search()
+            if inText:
+                page.append(line)
+            continue
+        m = tagRE.search(line)
+        if not m:
+            continue
+        tag = m.group(2)
+        if tag == 'page':
+            page = []
+            redirect = False
+        elif tag == 'id' and not id:
+            id = m.group(3)
+        elif tag == 'id' and id: # <revision> <id></id> </revision>
+            revid = m.group(3)
+        elif tag == 'title':
+            title = m.group(3)
+        elif tag == 'redirect':
+            redirect = True
+        elif tag == 'text':
+            inText = True
+            line = line[m.start(3):m.end(3)]
+            page.append(line)
+            if m.lastindex == 4:  # open-close
+                inText = False
+        elif tag == '/text':
+            if m.group(1):
+                page.append(m.group(1))
+            inText = False
+        elif inText:
+            page.append(line)
+        elif tag == '/page':
+            colon = title.find(':')
+            if (colon < 0 or (title[:colon] in acceptedNamespaces) and id != last_id and
+                    not redirect and not title.startswith(templateNamespace)):
+                yield (id, revid, title, page)
+                last_id = id
+            id = ''
+            revid = ''
+            page = []
+            inText = False
+            redirect = False
+
+
 def process_dump(input_file, template_file, out_file, file_size, file_compress,
-                 process_count):
+                 process_count, html_safe, expand_templates=True):
     """
     :param input_file: name of the wikipedia dump file; '-' to read from stdin
     :param template_file: optional file with template definitions.
@@ -282,12 +344,15 @@ def process_dump(input_file, template_file, out_file, file_size, file_compress,
     :param file_size: max size of each extracted file, or None for no max (one file)
     :param file_compress: whether to compress files with bzip.
     :param process_count: number of extraction processes to spawn.
+    :html_safe: whether to convert entities in text to HTML.
+    :param expand_templates: whether to expand templates.
     """
-    global urlbase
     global knownNamespaces
-    global templateNamespace, templatePrefix
+    global templateNamespace
     global moduleNamespace, modulePrefix
 
+    urlbase = ''                # This is obtained from <siteinfo>
+
     input = decode_open(input_file)
 
     # collect siteinfo
@@ -306,7 +371,7 @@ def process_dump(input_file, template_file, out_file, file_size, file_compress,
             knownNamespaces.add(m.group(3))
             if re.search('key="10"', line):
                 templateNamespace = m.group(3)
-                templatePrefix = templateNamespace + ':'
+                Extractor.templatePrefix = templateNamespace + ':'
             elif re.search('key="828"', line):
                 moduleNamespace = m.group(3)
                 modulePrefix = moduleNamespace + ':'
@@ -348,6 +413,9 @@ def process_dump(input_file, template_file, out_file, file_size, file_compress,
     # - pages to be processed are dispatched to workers
     # - a reduce process collects the results, sort them and print them.
 
+    # fixes MacOS error: TypeError: cannot pickle '_io.TextIOWrapper' object
+    Process = get_context("fork").Process
+
     maxsize = 10 * process_count
     # output queue
     output_queue = Queue(maxsize=maxsize)
@@ -364,7 +432,7 @@ def process_dump(input_file, template_file, out_file, file_size, file_compress,
     workers = []
     for _ in range(max(1, process_count)):
         extractor = Process(target=extract_process,
-                            args=(jobs_queue, output_queue))
+                            args=(jobs_queue, output_queue, html_safe))
         extractor.daemon = True  # only live while parent process lives
         extractor.start()
         workers.append(extractor)
@@ -373,53 +441,12 @@ def process_dump(input_file, template_file, out_file, file_size, file_compress,
 
     # we collect individual lines, since str.join() is significantly faster
     # than concatenation
-    page = []
-    id = None
-    last_id = None
+
     ordinal = 0  # page count
-    inText = False
-    redirect = False
-    for line in input:
-        #line = line.decode('utf-8')
-        if '<' not in line:  # faster than doing re.search()
-            if inText:
-                page.append(line)
-            continue
-        m = tagRE.search(line)
-        if not m:
-            continue
-        tag = m.group(2)
-        if tag == 'page':
-            page = []
-            redirect = False
-        elif tag == 'id' and not id:
-            id = m.group(3)
-        elif tag == 'title':
-            title = m.group(3)
-        elif tag == 'redirect':
-            redirect = True
-        elif tag == 'text':
-            inText = True
-            line = line[m.start(3):m.end(3)]
-            page.append(line)
-            if m.lastindex == 4:  # open-close
-                inText = False
-        elif tag == '/text':
-            if m.group(1):
-                page.append(m.group(1))
-            inText = False
-        elif inText:
-            page.append(line)
-        elif tag == '/page':
-            colon = title.find(':')
-            if (colon < 0 or (title[:colon] in acceptedNamespaces) and id != last_id and
-                    not redirect and not title.startswith(templateNamespace)):
-                job = (id, title, page, ordinal)
-                jobs_queue.put(job)  # goes to any available extract_process
-                last_id = id
-                ordinal += 1
-            id = None
-            page = []
+    for id, revid, title, page in collect_pages(input):
+        job = (id, revid, urlbase, title, page, ordinal)
+        jobs_queue.put(job)  # goes to any available extract_process
+        ordinal += 1
 
     input.close()
 
@@ -447,25 +474,27 @@ def process_dump(input_file, template_file, out_file, file_size, file_compress,
 # Multiprocess support
 
 
-def extract_process(jobs_queue, output_queue):
+def extract_process(jobs_queue, output_queue, html_safe):
     """Pull tuples of raw page content, do CPU/regex-heavy fixup, push finished text
     :param jobs_queue: where to get jobs.
     :param output_queue: where to queue extracted text for output.
+    :html_safe: whether to convert entities in text to HTML.
     """
     while True:
-        job = jobs_queue.get()  # job is (id, title, page, ordinal)
+        job = jobs_queue.get()  # job is (id, revid, urlbase, title, page)
         if job:
             out = StringIO()  # memory buffer
-            Extractor(*job[:3]).extract(out)  # (id, title, page)
+            Extractor(*job[:-1]).extract(out, html_safe)  # (id, urlbase, title, page)
             text = out.getvalue()
-            output_queue.put((job[3], text))  # (ordinal, extracted_text)
+            output_queue.put((job[-1], text))  # (ordinal, extracted_text)
             out.close()
         else:
             break
 
 
 def reduce_process(output_queue, output):
-    """Pull finished article text, write series of files (or stdout)
+    """
+    Pull finished article text, write series of files (or stdout)
     :param output_queue: text to be output.
     :param output: file object where to print.
     """
@@ -501,8 +530,8 @@ minFileSize = 200 * 1024
 
 
 def main():
-    global urlbase, acceptedNamespaces
-    global expand_templates, templateCache, escape_doc
+    global acceptedNamespaces
+    global templateCache
 
     parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]),
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
@@ -513,10 +542,12 @@ def main():
     groupO.add_argument("-o", "--output", default="text",
                         help="directory for extracted files (or '-' for dumping to stdout)")
     groupO.add_argument("-b", "--bytes", default="1M",
-                        help="maximum bytes per output file (default %(default)s)",
+                        help="maximum bytes per output file (default %(default)s); 0 means to put a single article per file",
                         metavar="n[KMG]")
     groupO.add_argument("-c", "--compress", action="store_true",
                         help="compress output files using bzip")
+    groupO.add_argument("--json", action="store_true",
+                        help="write output in json format instead of the default <doc> format")
 
     groupP = parser.add_argument_group('Processing')
     groupP.add_argument("--html", action="store_true",
@@ -527,10 +558,10 @@ def main():
                         help="accepted namespaces")
     groupP.add_argument("--templates",
                         help="use or create file containing templates")
-    groupP.add_argument("--no-templates", action="store_false",
+    groupP.add_argument("--no-templates", action="store_true",
                         help="Do not expand templates")
-    groupP.add_argument("--escapedoc", action="store_true",
-                        help="use to escape the contents of the output <doc>...</doc>")
+    groupP.add_argument("--html-safe", default=True,
+                        help="use to produce HTML safe output within <doc>...</doc>")
     default_process_count = cpu_count() - 1
     parser.add_argument("--processes", type=int, default=default_process_count,
                         help="Number of processes to use (default %(default)s)")
@@ -552,14 +583,13 @@ def main():
     Extractor.HtmlFormatting = args.html
     if args.html:
         Extractor.keepLinks = True
-
-    expand_templates = args.no_templates
-    escape_doc = args.escapedoc
+    Extractor.to_json = args.json
 
     try:
         power = 'kmg'.find(args.bytes[-1].lower()) + 1
-        file_size = int(args.bytes[:-1]) * 1024 ** power
-        if file_size < minFileSize:
+        # 0 bytes means put a single article per file.
+        file_size = 0 if args.bytes == '0' else int(args.bytes[:-1]) * 1024 ** power
+        if file_size and file_size < minFileSize:
             raise ValueError()
     except ValueError:
         logging.error('Insufficient or invalid size: %s', args.bytes)
@@ -592,17 +622,10 @@ def main():
                 with open(args.templates) as file:
                     load_templates(file)
 
-        with open(input_file) as file:
-            page = file.read()#.decode('utf-8')
-            m = re.search(r'<id>(.*)</id>', page)
-            id = m.group(1) if m else 0
-            m = re.search(r'<title>(.*)</title>', page)
-            if m:
-                title = m.group(1)
-            else:
-                logging.error('Missing title element')
-                return
-            Extractor(id, title, [page]).extract(sys.stdout)
+        urlbase = ''
+        with open(input_file) as input:
+            for id, revid, title, page in collect_pages(input):
+                Extractor(id, revid, urlbase, title, page).extract(sys.stdout)
         return
 
     output_path = args.output
@@ -614,8 +637,7 @@ def main():
             return
 
     process_dump(input_file, args.templates, output_path, file_size,
-                 args.compress, args.processes)
-
+                 args.compress, args.processes, args.html_safe, not args.no_templates)
 
 if __name__ == '__main__':
     main()
