@@ -20,8 +20,9 @@
 
 import re
 import html
+import json
 from itertools import zip_longest
-import urllib
+from urllib.parse import quote as urlencode
 from html.entities import name2codepoint
 import logging
 import time
@@ -34,9 +35,6 @@ syntaxhighlight = re.compile('&lt;syntaxhighlight .*?&gt;(.*?)&lt;/syntaxhighlig
 
 ## PARAMS ####################################################################
 
-# This is obtained from <siteinfo>
-urlbase = ''
-
 ##
 # Defined in <siteinfo>
 # We include as default Template, when loading external template file.
@@ -62,21 +60,21 @@ discardElements = [
 acceptedNamespaces = ['w', 'wiktionary', 'wikt']
 
 
-def get_url(uid):
+def get_url(urlbase, uid):
     return "%s?curid=%s" % (urlbase, uid)
 
 
 # ======================================================================
 
 
-def clean(extractor, text, expand_templates=False, escape_doc=True):
+def clean(extractor, text, expand_templates=False, html_safe=True):
     """
     Transforms wiki markup. If the command line flag --escapedoc is set then the text is also escaped
     @see https://www.mediawiki.org/wiki/Help:Formatting
     :param extractor: the Extractor t use.
     :param text: the text to clean.
     :param expand_templates: whether to perform template expansion.
-    :param escape_doc: whether to convert special characters to HTML entities.
+    :param html_safe: whether to convert reserved HTML characters to entities.
     @return: the cleaned text.
     """
 
@@ -174,7 +172,7 @@ def clean(extractor, text, expand_templates=False, escape_doc=True):
     text = re.sub(u'(\[\(«) ', r'\1', text)
     text = re.sub(r'\n\W+?\n', '\n', text, flags=re.U)  # lines with only punctuations
     text = text.replace(',,', ',').replace(',.', '.')
-    if escape_doc:
+    if html_safe:
         text = html.escape(text, quote=False)
     return text
 
@@ -201,7 +199,12 @@ def compact(text, mark_headers=False):
     for line in text.split('\n'):
 
         if not line:
+            if len(listLevel):    # implies Extractor.HtmlFormatting
+                for c in reversed(listLevel):
+                    page.append(listClose[c])
+                    listLevel = ''
             continue
+
         # Handle section titles
         m = section.match(line)
         if m:
@@ -229,36 +232,35 @@ def compact(text, mark_headers=False):
                 page.append(title)
         # handle indents
         elif line[0] == ':':
-            # page.append(line.lstrip(':*#;'))
-            continue
+            page.append(line.lstrip(':'))
         # handle lists
-        elif line[0] in '*#;:':
+        # @see https://www.mediawiki.org/wiki/Help:Formatting
+        elif line[0] in '*#;':
             if Extractor.HtmlFormatting:
-                i = 0
-                for c, n in zip_longest(listLevel, line, fillvalue=''):
-                    if not n or n not in '*#;:':
-                        if c:
-                            page.append(listClose[c])
-                            listLevel = listLevel[:-1]
-                            continue
-                        else:
-                            break
-                    # n != ''
-                    if c != n and (not c or (c not in ';:' and n not in ';:')):
-                        if c:
-                            # close level
-                            page.append(listClose[c])
-                            listLevel = listLevel[:-1]
-                        listLevel += n
-                        page.append(listOpen[n])
-                    i += 1
-                n = line[i - 1]  # last list char
-                line = line[i:].strip()
-                if line:  # FIXME: n is '"'
-                    page.append(listItem[n] % line)
+                # close extra levels
+                l = 0
+                for c in listLevel:
+                    if l < len(line) and c != line[l]:
+                        for extra in reversed(listLevel[l:]):
+                            page.append(listClose[extra])
+                        listLevel = listLevel[:l]
+                        break
+                    l += 1
+                if l < len(line) and line[l] in '*#;:':
+                    # add new level (only one, no jumps)
+                    # FIXME: handle jumping levels
+                    type = line[l]
+                    page.append(listOpen[type])
+                    listLevel += type
+                    line = line[l+1:].strip()
+                else:
+                    # continue on same level
+                    type = line[l-1]
+                    line = line[l:].strip()
+                page.append(listItem[type] % line)
             else:
                 continue
-        elif len(listLevel):
+        elif len(listLevel):    # implies Extractor.HtmlFormatting
             for c in reversed(listLevel):
                 page.append(listClose[c])
             listLevel = []
@@ -422,7 +424,7 @@ def replaceExternalLinks(text):
 def makeExternalLink(url, anchor):
     """Function applied to wikiLinks"""
     if Extractor.keepLinks:
-        return '<a href="%s">%s</a>' % (urllib.quote(url.encode('utf-8')), anchor)
+        return '<a href="%s">%s</a>' % (urlencode(url), anchor)
     else:
         return anchor
 
@@ -492,7 +494,7 @@ def makeInternalLink(title, label):
         if colon2 > 1 and title[colon + 1:colon2] not in acceptedNamespaces:
             return ''
     if Extractor.keepLinks:
-        return '<a href="%s">%s</a>' % (urllib.quote(title), label)
+        return '<a href="%s">%s</a>' % (urlencode(title), label)
     else:
         return label
 
@@ -788,6 +790,114 @@ spaces = re.compile(r' {2,}')
 # Matches dots
 dots = re.compile(r'\.{4,}')
 
+# ======================================================================
+
+class Template(list):
+    """
+    A Template is a list of TemplateText or TemplateArgs
+    """
+
+    @classmethod
+    def parse(cls, body):
+        tpl = Template()
+        # we must handle nesting, s.a.
+        # {{{1|{{PAGENAME}}}
+        # {{{italics|{{{italic|}}}
+        # {{#if:{{{{{#if:{{{nominee|}}}|nominee|candidate}}|}}}|
+        #
+        start = 0
+        for s,e in findMatchingBraces(body, 3):
+            tpl.append(TemplateText(body[start:s]))
+            tpl.append(TemplateArg(body[s+3:e-3]))
+            start = e
+        tpl.append(TemplateText(body[start:])) # leftover
+        return tpl
+
+    def subst(self, params, extractor, depth=0):
+        # We perform parameter substitutions recursively.
+        # We also limit the maximum number of iterations to avoid too long or
+        # even endless loops (in case of malformed input).
+
+        # :see: http://meta.wikimedia.org/wiki/Help:Expansion#Distinction_between_variables.2C_parser_functions.2C_and_templates
+        #
+        # Parameter values are assigned to parameters in two (?) passes.
+        # Therefore a parameter name in a template can depend on the value of
+        # another parameter of the same template, regardless of the order in
+        # which they are specified in the template call, for example, using
+        # Template:ppp containing "{{{{{{p}}}}}}", {{ppp|p=q|q=r}} and even
+        # {{ppp|q=r|p=q}} gives r, but using Template:tvvv containing
+        # "{{{{{{{{{p}}}}}}}}}", {{tvvv|p=q|q=r|r=s}} gives s.
+
+        logging.debug('subst tpl (%d, %d) %s', len(extractor.frame), depth, self)
+
+        if depth > extractor.maxParameterRecursionLevels:
+            extractor.recursion_exceeded_3_errs += 1
+            return ''
+
+        return ''.join([tpl.subst(params, extractor, depth) for tpl in self])
+
+    def __str__(self):
+        return ''.join([str(x) for x in self])
+
+
+class TemplateText(str):
+    """Fixed text of template"""
+
+    def subst(self, params, extractor, depth):
+        return self
+
+
+class TemplateArg():
+    """
+    parameter to a template.
+    Has a name and a default value, both of which are Templates.
+    """
+    def __init__(self, parameter):
+        """
+        :param parameter: the parts of a tplarg.
+        """
+        # the parameter name itself might contain templates, e.g.:
+        #   appointe{{#if:{{{appointer14|}}}|r|d}}14|
+        #   4|{{{{{subst|}}}CURRENTYEAR}}
+
+        # any parts in a tplarg after the first (the parameter default) are
+        # ignored, and an equals sign in the first part is treated as plain text.
+        #logging.debug('TemplateArg %s', parameter)
+
+        parts = splitParts(parameter)
+        self.name = Template.parse(parts[0])
+        if len(parts) > 1:
+            # This parameter has a default value
+            self.default = Template.parse(parts[1])
+        else:
+            self.default = None
+
+    def __str__(self):
+        if self.default:
+            return '{{{%s|%s}}}' % (self.name, self.default)
+        else:
+            return '{{{%s}}}' % self.name
+
+    def subst(self, params, extractor, depth):
+        """
+        Substitute value for this argument from dict :param params:
+        Use :param extractor: to evaluate expressions for name and default.
+        Limit substitution to the maximun :param depth:.
+        """
+        # the parameter name itself might contain templates, e.g.:
+        # appointe{{#if:{{{appointer14|}}}|r|d}}14|
+        paramName = self.name.subst(params, extractor, depth+1)
+        paramName = extractor.expandTemplates(paramName)
+        res = ''
+        if paramName in params:
+            res = params[paramName]  # use parameter value specified in template invocation
+        elif self.default:            # use the default value
+            defaultValue = self.default.subst(params, extractor, depth+1)
+            res =  extractor.expandTemplates(defaultValue)
+        #logging.debug('subst arg %d %s -> %s' % (depth, paramName, res))
+        return res
+
+
 # ======================================================================
 
 substWords = 'subst:|safesubst:'
@@ -809,11 +919,21 @@ class Extractor():
     # Whether to output text with HTML formatting elements in <doc> files.
     HtmlFormatting = False
 
-    def __init__(self, id, title, page):
+    ##
+    # Whether to produce json instead of the default <doc> output format.
+    toJson = False
+
+    ##
+    # Obtained from TemplateNamespace
+    templatePrefix = ''
+
+    def __init__(self, id, revid, urlbase, title, page):
         """
         :param page: a list of lines.
         """
         self.id = id
+        self.revid = revid
+        self.url = get_url(urlbase, id)
         self.title = title
         self.page = page
         self.magicWords = MagicWords()
@@ -823,12 +943,14 @@ class Extractor():
         self.recursion_exceeded_3_errs = 0  # parameter recursion
         self.template_title_errs = 0
 
-    def clean_text(self, text, mark_headers=False, expand_templates=False,
-                   escape_doc=True):
+    def clean_text(self, text, mark_headers=False, expand_templates=True,
+                   html_safe=True):
         """
         :param mark_headers: True to distinguish headers from paragraphs
           e.g. "## Section 1"
         """
+        self.magicWords['namespace'] = self.title[:max(0, self.title.find(":"))]
+        #self.magicWords['namespacenumber'] = '0' # for article, 
         self.magicWords['pagename'] = self.title
         self.magicWords['fullpagename'] = self.title
         self.magicWords['currentyear'] = time.strftime('%Y')
@@ -838,31 +960,41 @@ class Extractor():
         self.magicWords['currenttime'] = time.strftime('%H:%M:%S')
 
         text = clean(self, text, expand_templates=expand_templates,
-                     escape_doc=escape_doc)
+                     html_safe=html_safe)
 
         text = compact(text, mark_headers=mark_headers)
         return text
 
-    def extract(self, out, escape_doc=True):
+    def extract(self, out, html_safe=True):
         """
         :param out: a memory file.
+        :param html_safe: whether to escape HTML entities.
         """
         logging.debug("%s\t%s", self.id, self.title)
         text = ''.join(self.page)
+        text = self.clean_text(text, html_safe=html_safe)
 
-        url = get_url(self.id)
-        header = '<doc id="%s" url="%s" title="%s">\n' % (self.id, url, self.title)
-        # Separate header from text with a newline.
-        header += self.title + '\n\n'
-        footer = "\n</doc>\n"
-        out.write(header)
-
-        text = self.clean_text(text, escape_doc=escape_doc)
-
-        for line in text:
-            out.write(line)
+        if self.to_json:
+            json_data = {
+		'id': self.id,
+                'revid': self.revid,
+                'url': self.url,
+                'title': self.title,
+                'text': "\n".join(text)
+            }
+            out_str = json.dumps(json_data)
+            out.write(out_str)
             out.write('\n')
-        out.write(footer)
+        else:
+            header = '<doc id="%s" url="%s" title="%s">\n' % (self.id, self.url, self.title)
+            # Separate header from text with a newline.
+            header += self.title + '\n\n'
+            footer = "\n</doc>\n"
+            out.write(header)
+            out.write('\n'.join(text))
+            out.write('\n')
+            out.write(footer)
+
         errs = (self.template_title_errs,
                 self.recursion_exceeded_1_errs,
                 self.recursion_exceeded_2_errs,
@@ -875,7 +1007,7 @@ class Extractor():
     # Expand templates
 
     maxTemplateRecursionLevels = 30
-    maxParameterRecursionLevels = 10
+    maxParameterRecursionLevels = 16
 
     # check for template beginning
     reOpen = re.compile('(?<!{){{(?!{)', re.DOTALL)
@@ -964,7 +1096,11 @@ class Extractor():
             # The '=' might occurr within an HTML attribute:
             #   "&lt;ref name=value"
             # but we stop at first.
-            m = re.match(' *([^=]*?) *=(.*)', param, re.DOTALL)
+
+            # The '=' might occurr within quotes:
+            # ''''<span lang="pt-pt" xml:lang="pt-pt">cénicas</span>'''
+
+            m = re.match(" *([^=']*?) *=(.*)", param, re.DOTALL)
             if m:
                 # This is a named parameter.  This case also handles parameter
                 # assignments like "2=xxx", where the number of an unnamed
@@ -1259,7 +1395,7 @@ def findMatchingBraces(text, ldelim=0):
 
     if ldelim:  # 2-3
         reOpen = re.compile('[{]{%d,}' % ldelim)  # at least ldelim
-        reNext = re.compile('[{]{2,}|}{2,}')  # at least 2
+        reNext = re.compile('[{]{2,}|}{2,}')  # at least 2 open or close bracces
     else:
         reOpen = re.compile('{{2,}|\[{2,}')
         reNext = re.compile('{{2,}|}{2,}|\[{2,}|]{2,}')  # at least 2
@@ -1425,7 +1561,7 @@ def fullyQualifiedTemplateTitle(templateTitle):
     # space]], but having in the system a redirect page with an empty title
     # causes numerous problems, so we'll live happier without it.
     if templateTitle:
-        return templatePrefix + ucfirst(templateTitle)
+        return Extractor.templatePrefix + ucfirst(templateTitle)
     else:
         return ''  # caller may log as error
 
@@ -1475,7 +1611,7 @@ def sharp_expr(expr):
         expr = re.sub('mod', '%', expr)
         expr = re.sub('\bdiv\b', '/', expr)
         expr = re.sub('\bround\b', '|ROUND|', expr)
-        return unicode(eval(expr))
+        return str(eval(expr))
     except:
         return '<span class="error"></span>'
 
@@ -1615,7 +1751,7 @@ parserFunctions = {
 
     # This function is used in some pages to construct links
     # http://meta.wikimedia.org/wiki/Help:URL
-    'urlencode': lambda string, *rest: urllib.quote(string.encode('utf-8')),
+    'urlencode': lambda string, *rest: urlencode(string),
 
     'lc': lambda string, *rest: string.lower() if string else '',
 
@@ -1627,6 +1763,8 @@ parserFunctions = {
 
     'int': lambda string, *rest: str(int(string)),
 
+    'padleft': lambda char, width, string: string.ljust(char, int(pad)), # CHECK_ME
+
 }
 
 
@@ -1634,6 +1772,8 @@ def callParserFunction(functionName, args, frame):
     """
     Parser functions have similar syntax as templates, except that
     the first argument is everything after the first colon.
+    :param functionName: nameof the parser function
+    :param args: the arguments to the function
     :return: the result of the invocation, None in case of failure.
 
     http://meta.wikimedia.org/wiki/Help:ParserFunctions
@@ -1643,11 +1783,11 @@ def callParserFunction(functionName, args, frame):
         if functionName == '#invoke':
             # special handling of frame
             ret = sharp_invoke(args[0].strip(), args[1].strip(), frame)
-            # logging.debug('parserFunction> %s %s', functionName, ret)
+            # logging.debug('parserFunction> %s %s', args[1], ret)
             return ret
         if functionName in parserFunctions:
             ret = parserFunctions[functionName](*args)
-            # logging.debug('parserFunction> %s %s', functionName, ret)
+            # logging.debug('parserFunction> %s(%s) %s', functionName, args, ret)
             return ret
     except:
         return ""  # FIXME: fix errors
@@ -1661,7 +1801,7 @@ def callParserFunction(functionName, args, frame):
 reNoinclude = re.compile(r'<noinclude>(?:.*?)</noinclude>', re.DOTALL)
 reIncludeonly = re.compile(r'<includeonly>|</includeonly>', re.DOTALL)
 
-# These are built before spawning processes, hence thay are shared.
+# These are built before spawning processes, hence they are shared.
 templates = {}
 redirects = {}
 # cache of parser templates
@@ -1714,6 +1854,6 @@ def define_template(title, page):
         text = reIncludeonly.sub('', text)
 
     if text:
-        if title in templates:
+        if title in templates and templates[title] != text:
             logging.warn('Redefining: %s', title)
         templates[title] = text
