@@ -34,7 +34,7 @@ import bz2
 
 
 # Program version
-__version__ = '3.0.4'
+__version__ = '3.0.5'
 
 # ----------------------------------------------------------------------
 # READER
@@ -46,7 +46,8 @@ tagRE = re.compile(r'(.*?)<(/?\w+)[^>]*>(?:([^<]*)(<.*?>)?)?')
 def process_data(input_file, id, templates=False):
     """
     :param input_file: name of the wikipedia dump file.
-    :param id: article id
+    :param id: article id.
+    :param templates: whether article is a template.
     """
 
     if input_file.lower().endswith(".bz2"):
@@ -105,9 +106,9 @@ def main():
     parser.add_argument("--id", default="1",
                         help="article number")
     parser.add_argument("--template", action="store_true",
-                        help="template number")
+                        help="whether article is a template")
     parser.add_argument("-v", "--version", action="version",
-                        version='%(prog)s ' + version,
+                        version='%(prog)s ' + __version__,
                         help="print program version")
 
     args = parser.parse_args()
